"""Memcached functioanl tests."""

import logging
import telnetlib
import time
import unittest

# import requests
#
# from zaza.utilities import juju
from zaza import model


# Create a configuration dictionary for custom memcached values.
CONFIGURATION = {
    "size": 768,
    "connection-limit": 1024,
    "factor": 1.25,
    "tcp-port": 11214,
    "udp-port": 11213,
}
DEFAULT_PORT = 11211
DEFAULT_TIMEOUT = 30


class BaseMemcachedTest(unittest.TestCase):
    """Base for Memcached charm tests."""

    @classmethod
    def setUpClass(cls):
        """Set common value."""
        super(BaseMemcachedTest, cls).setUpClass()
        cls.model_name = model.get_juju_model()
        cls.application_name = "memcached"
        cls.lead_unit_name = model.get_lead_unit_name(
            cls.application_name, model_name=cls.model_name
        )
        cls.units = model.get_units(cls.application_name, model_name=cls.model_name)
        logging.debug("Lead unit is {}".format(cls.lead_unit_name))
        cls.memcached_ip = model.get_app_ips(cls.application_name)[0]

        cls.config = model.get_application_config(cls.application_name)

        # open memcache to be able to connect from this machine
        cmd = "ufw insert 1 allow {port}".format(port=DEFAULT_PORT)
        for unit in cls.units:
            response = model.run_on_unit(unit.name, cmd)
            if response["Code"] != "0":
                cls.fail(
                    "Could not open port {port} for any source ({unit})".format(
                        port=DEFAULT_PORT, unit=unit.name
                    )
                )


class CharmOperationTest(BaseMemcachedTest):
    """Verify operations."""

    def test_service_up(self):
        """Check if the service is up."""
        cmd = (
            "/usr/local/lib/nagios/plugins/check_memcache.py"
            " -H 127.0.0.1 -p 11211 -c version"
        )
        timeout = time.time() + DEFAULT_TIMEOUT
        while time.time() < timeout:
            response = model.run_on_unit(self.lead_unit_name, cmd)
            if response["Code"] == "0":
                return
            logging.warning(
                "Unexpected nrpe check response: {}. Retrying in 5s."
                "".format(response)
            )
            time.sleep(5)

        self.fail(
            "Memcached didn't respond to the command {cmd} as expected.\n"
            "Result: {result}".format(cmd=cmd, result=response)
        )

    def test_memcached_store_value(self):
        """Store value test."""
        try:
            # Connect to memcached via telnet.
            tn = telnetlib.Telnet(self.memcached_ip, DEFAULT_PORT)
            date_time = time.strftime("%F %r")
            # Write the command that gets the current greeting.
            tn.write(b"get greeting\r\n")
            # Read the current greeting.
            response = tn.read_until(b"END", 2)
            # Create a string with date and time for this test.
            string = "memcached test {}".format(date_time)
            command = "set greeting 1 0 {}".format(len(string))
            # Write the command that sets the new greeting.
            tn.write(command.encode() + b"\r\n")
            tn.write(string.encode() + b"\r\n")
            # Read the response
            response = tn.read_until(b"STORED", 2)
            # Write the command that gets the greeting.
            tn.write(b"get greeting\r\n")
            # Get the new greeting in memcached.
            response = tn.read_until(b"END", 2)
            response = response.decode()
            logging.debug("get greeting response: {}".format(response))
            # Look for the string in the response.
            index = response.find(string)
            if index == -1:
                message = "Did not find {} in the greeting from memcached.".format(
                    string
                )
                self.fail(message)

            logging.info("Found {} in the greeting response.".format(string))
            tn.write(b"quit\n")
        except Exception as e:
            message = (
                "An error occurred communicating with memcached over telnet "
                "{0}:{1} {2}"
            ).format(self.memcached_ip, DEFAULT_PORT, str(e))
            self.fail(message)
        finally:
            try:
                tn.close()
            except NameError:
                pass

    def test_memcached_file_content(self):
        """Test if the memcached service is configured properly."""
        # Get the contents of the memcached configuration file.
        response = model.run_on_unit(self.lead_unit_name, "cat /etc/memcached.conf")
        if response["Code"] != "0":
            self.fail("cmd failed: {}".format(response))
        config_string = response["Stdout"]

        # Parse the configuration file for the values sent in the deployment.
        remote_config = {}
        for line in config_string.splitlines():
            if line.startswith("-m "):
                remote_config["size"] = line.split()[1]
            elif line.startswith("-c "):
                remote_config["connection-limit"] = line.split()[1]
            elif line.startswith("-f "):
                remote_config["factor"] = line.split()[1]

        # Check for the configured values.
        if any(
            float(CONFIGURATION[opt]) != float(remote_config.get(opt))
            for opt in ("size", "connection-limit", "factor")
        ):
            message = (
                "The memcached deployment was not configured correctly, size: "
                "{size} limit: {connection-limit} factor: {factor}"
                "".format(**remote_config)
            )
            self.fail(message)

        logging.info("The memcached deployment was configured correctly.")

    def test_memcached_check_replication(self):
        """Test if the memcached replication is configured properly."""
        repcached = self.config["repcached"]["value"]
        if not repcached:
            logging.info("test_memcached_check_replication skipped")
            return

        date_time = time.strftime("%F %r")
        string = "memcached test %s" % date_time

        try:
            # Connect to memcached via telnet.
            tn = telnetlib.Telnet(self.memcached_ip, DEFAULT_PORT)
            command = "set greeting 1 0 %d" % len(string)
            tn.write(command.encode() + b"\r\n")
            tn.write(string.encode() + b"\r\n")
            response = tn.read_until(b"STORED", 2)
        except Exception as e:
            message = (
                "An error occurred communicating with memcached over telnet "
                "{0}:{1} {2}"
            ).format(self.memcached_ip, DEFAULT_PORT, str(e))
            self.fail(message)
        finally:
            try:
                tn.close()
            except NameError:
                pass

        if repcached:
            # give time for replication
            time.sleep(2)

        for unit_ip in model.get_app_ips(self.application_name):
            try:
                # check stored value in all units
                tn = telnetlib.Telnet(unit_ip, DEFAULT_PORT)
                tn.write(b"get greeting\r\n")
                response = tn.read_until(b"END", 2)
                response = response.decode()
                index = response.find(string)

                if index == -1:
                    tn.write(b"quit\n")
                    tn.close()
                    message = (
                        "Did not find {} in the greeting from memcached:"
                        " unit[{}]".format(string, unit_ip)
                    )
                    self.fail(message)
                else:
                    logging.info(
                        "Found string[{}] in the greeting response (unit[{}])."
                        "".format(string, unit_ip)
                    )

                tn.write(b"quit\n")
            except Exception as e:
                message = (
                    "An error occurred communicating with memcached over telnet "
                    "{0}:{1} {2}"
                ).format(unit_ip, DEFAULT_PORT, str(e))
                self.fail(message)
            finally:
                try:
                    tn.close()
                except NameError:
                    pass
