# grabbed from nova-cloud-controller/unit_tests/test_utils.py
"""Utils tests."""

import io
import logging
import os
import unittest
from contextlib import contextmanager

from mock import MagicMock, patch

import yaml


def load_config():
    """Walk backwords from __file__ looking for config.yaml.

    Load and return the 'options' section'
    """
    config = None
    f = __file__
    while config is None:
        d = os.path.dirname(f)
        if os.path.isfile(os.path.join(d, "config.yaml")):
            config = os.path.join(d, "config.yaml")
            break
        f = d

    if not config:
        logging.error(
            "Could not find config.yaml in any parent directory of %s. " % __file__
        )
        raise Exception

    with open(config) as conf_file:
        return yaml.safe_load(conf_file.read())["options"]


def get_default_config():
    """Load default charm config from config.yaml return as a dict.

    If no default is set in config.yaml, its value is None.
    """
    default_config = {}
    config = load_config()
    for k, v in config.items():
        if "default" in v:
            default_config[k] = v["default"]
        else:
            default_config[k] = None
    return default_config


class CharmTestCase(unittest.TestCase):
    """Test charm."""

    def setUp(self, obj, patches):
        """Set common values."""
        super(CharmTestCase, self).setUp()
        self.patches = patches
        self.obj = obj
        self.test_config = TestConfig()
        self.test_relation = TestRelation()
        self.patch_all()

    def patch(self, method):
        """Patch one method."""
        _m = patch.object(self.obj, method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    def patch_all(self):
        """Patch all method."""
        for method in self.patches:
            setattr(self, method, self.patch(method))


class TestConfig(object):
    """Config test."""

    def __init__(self):
        """Init."""
        self.config = get_default_config()

    def get(self, attr=None):
        """Get value."""
        if not attr:
            return self.get_all()
        try:
            return self.config[attr]
        except KeyError:
            return None

    def get_all(self):
        """Get all config."""
        return self.config

    def set(self, attr, value):
        """Set value."""
        if attr not in self.config:
            raise KeyError
        self.config[attr] = value


class TestRelation(object):
    """Relation tests."""

    def __init__(self, relation_data={}):
        """Init."""
        self.relation_data = relation_data

    def set(self, relation_data):
        """Set value."""
        self.relation_data = relation_data

    def get(self, attr=None, unit=None, rid=None):
        """Get value."""
        if attr is None:
            return self.relation_data
        elif attr in self.relation_data:
            return self.relation_data[attr]
        return None


@contextmanager
def patch_open():
    """Patch open() to allow mocking open() and the file that is yielded.

    Yields the mock for "open" and "file", respectively.
    """
    mock_open = MagicMock(spec="builtins.open")
    mock_file = MagicMock(spec=io.FileIO)

    @contextmanager
    def stub_open(*args, **kwargs):
        mock_open(*args, **kwargs)
        yield mock_file

    with patch("builtins.open", stub_open):
        yield mock_open, mock_file
